try:
    from http import HTTPStatus
except ImportError:
    # Python 3.4
    from checkout_sdk.enums import HTTPStatus

import checkout_sdk as sdk
from checkout_sdk import ApiClient, HTTPMethod, Validator, PaymentStatus
from checkout_sdk.common import ResponseDTO
from checkout_sdk.vaults.responses import (
    Instrument
)


class InstrumentsClient(ApiClient):
    def create(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            # dictionary approach - everything else is ignored
            request = args[0]
        else:
            # parameter approach
            request = kwargs

        token = request['token']

        self._log_info('Instrument Create Token:{}'.format(
            token
        ))

        http_response = self._send_http_request(
            'instruments', HTTPMethod.POST, request)

        return Instrument(http_response)

    def get(self, source_id):
        self._log_info('{} - {}'.format('get'.capitalize(), source_id))

        Validator.validate_id(source_id)

        http_response = self._send_http_request(
            'instruments/{}'.format(source_id), HTTPMethod.GET)

        return Instrument(http_response)

    def update(self, source_id, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            # dictionary approach - everything else is ignored
            request = args[0]
        else:
            # parameter approach
            request = kwargs

        Validator.validate_id(source_id)

        http_response = self._send_http_request(
            'instruments/{}'.format(source_id), HTTPMethod.PATCH, request)

        return Instrument(http_response)



from checkout_sdk.payments.responses import Payment


class PaymentProcessed(Payment):
    def __init__(self, api_response):
        super().__init__(api_response, is_pending=False)

    @property
    def actions_link(self):
        return self.get_link('actions')

    @property
    def can_capture(self):
        return self.has_link('capture')

    @property
    def capture_link(self):
        return self.get_link('capture')

    @property
    def can_void(self):
        return self.has_link('void')

    @property
    def void_link(self):
        return self.get_link('void')

    @property
    def can_refund(self):
        return self.has_link('refund')

    @property
    def refund_link(self):
        return self.get_link('refund')



from checkout_sdk.payments.responses import Payment


class PaymentPending(Payment):
    def __init__(self, api_response):
        super().__init__(api_response, is_pending=True)

    @property
    def requires_redirect(self):
        return self.has_link('redirect')

    @property
    def redirect_link(self):
        return self.get_link('redirect')



import re
import os

from checkout_sdk import constants, logger

SECRET_KEY_REGEX = re.compile(
    r'^sk_(test_)?(\w{8})-(\w{4})-(\w{4})-(\w{4})-(\w{12})$', re.IGNORECASE)
PUBLIC_KEY_REGEX = re.compile(
    r'^pk_(test_)?(\w{8})-(\w{4})-(\w{4})-(\w{4})-(\w{12})$', re.IGNORECASE)
HTTP_PROTOCOL_REGEX = re.compile(r'^http(s)?', re.IGNORECASE)

API_BASE_URLS = {
    'sandbox': 'https://api.sandbox.checkout.com/',
    'production': 'https://api.checkout.com/'
}

# using lambda functions here so that
# env variables are read at the time of access
ENV_SETTINGS = {
    'secret_key':
        lambda: os.environ.get('CKO_SECRET_KEY'),
    'public_key':
        lambda: os.environ.get('CKO_PUBLIC_KEY'),
    'sandbox':
        lambda: os.environ.get('CKO_SANDBOX', 'true').lower() in ['true', '1']
}


class Config:
    def __init__(self, secret_key=None, sandbox=None,
                 timeout=constants.DEFAULT_TIMEOUT, api_base_url=None,
                 public_key=None):
        self.secret_key = \
            secret_key if secret_key else ENV_SETTINGS['secret_key']()
        self.public_key = \
            public_key if public_key else ENV_SETTINGS['public_key']()
        self.timeout = timeout
        # use custom api_base_url if local dev, integration testing, etc
        # else `sandbox` determines the URL
        if api_base_url:
            self.api_base_url = api_base_url
        else:
            sandbox = \
                sandbox if sandbox is not None else ENV_SETTINGS['sandbox']()
            self.api_base_url = API_BASE_URLS['sandbox'] if sandbox is True \
                else API_BASE_URLS['production']

        logger.info('%s - sk%s%s', self.api_base_url, '*' * 6,
                    self.secret_key[-6:])

    @property
    def secret_key(self):
        return self._secret_key

    @secret_key.setter
    def secret_key(self, value):
        if SECRET_KEY_REGEX.match(value or ''):
            self._secret_key = value
        else:
            raise ValueError('Invalid secret key.')

    @property
    def public_key(self):
        return self._public_key

    @public_key.setter
    def public_key(self, value):
        if PUBLIC_KEY_REGEX.match(value or ''):
            self._public_key = value
        else:
            raise ValueError('Invalid public key.')

    @property
    def api_base_url(self):
        return self._api_base_url

    @api_base_url.setter
    def api_base_url(self, value):
        self._api_base_url = value

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        try:
            self._timeout = int(value)
        except ValueError:
            raise ValueError(
                'Invalid timeout. Default is {} milliseconds'
                .format(constants.DEFAULT_TIMEOUT))



import pprint

from checkout_sdk import logger


class ApiClient:
    def __init__(self, http_client):
        self._http_client = http_client

    def _send_http_request(self, url, method, request=None, headers=None):
        response = self._http_client.send(url, method, request, headers)
        self._log_info(response)  # http status, elapsed
        self._log_debug('HTTP response:\n' + pprint.pformat(response.body))
        return response

    def _log_info(self, message, **kwargs):
        logger.info(message, **kwargs)

    def _log_debug(self, message, **kwargs):
        logger.debug(message, **kwargs)



import logging
import os
import sys

name = "checkout_sdk"

from checkout_sdk.enums import Currency, HTTPMethod, PaymentType, PaymentStatus

# defaults
default_currency = None
default_capture = True
default_payment_type = PaymentType.Regular
default_response_immutable = True

# logging
logger = logging.getLogger('cko')
formatter = logging.Formatter('[%(levelname)s] %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

LOGGING_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO
}
env_logging_level = str(os.environ.get('CKO_LOGGING', None)).lower()
if env_logging_level in LOGGING_LEVELS.keys():
    logger.setLevel(LOGGING_LEVELS[env_logging_level])

# class imports (order matters!)

from checkout_sdk import constants
from checkout_sdk.config import Config
from checkout_sdk.validator import Validator

from checkout_sdk.http_client import HTTPClient
from checkout_sdk.api_client import ApiClient
from checkout_sdk.checkout_api import CheckoutApi

def get_api(secret_key=None, sandbox=None,
            timeout=constants.DEFAULT_TIMEOUT, api_base_url=None, public_key=None):
    return CheckoutApi(**locals())


try:
    from http import HTTPStatus
except ImportError:
    # Python 3.4
    from checkout_sdk.enums import HTTPStatus

import checkout_sdk as sdk
from checkout_sdk import ApiClient, HTTPMethod, Validator, PaymentStatus
from checkout_sdk.common import ResponseDTO
from checkout_sdk.vaults.responses import Exchange


class ExchangeClient(ApiClient):
    def request(self, path='payments', *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            # dictionary approach - everything else is ignored
            request = args[0]
        else:
            # parameter approach
            request = kwargs
        # parameter approach

        headers = request.get('headers')
        request.pop('headers', None)

        http_response = self._send_http_request(
            'exchange/{}'.format(path), HTTPMethod.POST, request, headers)

        return Exchange(http_response)



from checkout_sdk.vaults.tokens_client import TokensClient
from checkout_sdk import Config, HTTPClient
from checkout_sdk.payments import PaymentsClient
from checkout_sdk.vaults import TokensClient

class CheckoutApi:
    def __init__(self, **kwargs):
        config = Config(**kwargs)
        http_client = HTTPClient(config)
        if(config.secret_key is not None):
            self.payments = PaymentsClient(http_client)
        if(config.public_key is not None):
            self.tokens = TokensClient(http_client)



from checkout_sdk.common import Resource


class Payment(Resource):
    def __init__(self, api_response, is_pending):
        super().__init__(api_response)
        self._is_pending = is_pending

    @property
    def is_pending(self):
        return self._is_pending





class CheckoutSdkError(Exception):
    def __init__(self, request_id=None, api_version=None,
                 http_status=None, error_type=None, error_codes=None,
                 elapsed=0):
        super().__init__()
        self.request_id = request_id
        self.api_version = api_version
        self.http_status = http_status
        self.error_type = error_type
        self.error_codes = error_codes
        self.elapsed = elapsed

    def __str__(self):
        return '{} - {}'.format(self.request_id, self.error_type)


class AuthenticationError(CheckoutSdkError):
    pass


class ValidationError(CheckoutSdkError):
    pass


class ResourceNotFoundError(CheckoutSdkError):
    pass


class NotAllowedError(CheckoutSdkError):
    pass


class TooManyRequestsError(CheckoutSdkError):
    pass


class ApiError(CheckoutSdkError):
    pass


class ApiTimeoutError(CheckoutSdkError):
    pass



import checkout_sdk


class ResponseDTO:
    def __init__(self, kvp=None,
                 read_only=checkout_sdk.default_response_immutable):
        if kvp is not None and not isinstance(kvp, dict):
            raise TypeError('Dictionary expected')

        # load the DTO
        self._read_only = False
        self._kvp = {}
        if kvp is not None:
            for (k, v) in kvp.items():
                # calls __setitem__ further below
                self[k] = v

        # make the DTO immutable if required
        self._read_only = read_only

    # only called for attributes which are not declared directly
    def __getattr__(self, k):
        # this will in turn call __getitem__
        # we catch KeyError to raise AttributeError
        # this maintains expected behaviour
        #   when used with getattr() and hasattr()
        try:
            return self[k]
        except KeyError:
            raise AttributeError

    # KeyError if key does not exist
    def __getitem__(self, k):
        return self._kvp[k]

    # safer option with fallback option (None as default)
    def get(self, k, d=None):
        return self._kvp.get(k, d)

    def __setattr__(self, k, v):
        # avoiding infinite loop here
        # private attributes are registered via the super class (`object`)
        if k[0] == '_':
            super().__setattr__(k, v)
        else:
            # this will in turn call __setitem__
            try:
                self[k] = v
            except KeyError as err:
                raise AttributeError(err)

    def __setitem__(self, k, v):
        if self._read_only:
            raise KeyError(
                '{} is read-only.'.format(self.__class__.__name__))
        else:
            if isinstance(v, dict):
                self._kvp[k] = ResponseDTO(v, self._read_only)
            elif isinstance(v, list):
                self._kvp[k] = [ResponseDTO(item, self._read_only)
                                for item in v]
            else:
                self._kvp[k] = v

    def __len__(self):
        return len(self._kvp)



from checkout_sdk.payments.responses.payment import Payment

from checkout_sdk.payments.responses.payment_processed import PaymentProcessed
from checkout_sdk.payments.responses.payment_pending import PaymentPending
from checkout_sdk.payments.responses.payment_action import PaymentAction



import time
from urllib.parse import urljoin
import requests

from checkout_sdk import errors, constants, HTTPMethod
from checkout_sdk.common import HTTPResponse

HTTP_HEADERS_DEFAULTS = {
    'user-agent': 'checkout-sdk-python/{}'.format(constants.VERSION)
}


class HTTPClient:
    def __init__(self, config):
        self._config = config

        self._headers = HTTP_HEADERS_DEFAULTS.copy()

        # init HTTP Session (for pooling)
        self._session = requests.Session()

        # interceptor call is a mirror by default
        self.interceptor = lambda url, headers, request: (
            url, headers, request)

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = value

    def close_session(self):
        self._session.close()

    def send(self, path, method=HTTPMethod.GET, request=None, headers=None):
        start = time.time()
        if headers is not None:
            req_headers = headers
        else:
            req_headers = self.headers
            if 'tokens' in path:
                req_headers['authorization'] = self.config.public_key
            else:
                req_headers['authorization'] = self.config.secret_key

        # call the interceptor as a hook to override
        # the url, headers and/or request
        url, headers, request = self.interceptor(
            urljoin(self.config.api_base_url, path), req_headers, request)
        try:
            response = self._session.request(
                method=method,
                url=url,
                json=request,
                headers=headers,
                timeout=self.config.timeout / 1000)
            elapsed = self._calc_elapsed_time(start)

            response.raise_for_status()
            body = response.json()

            return HTTPResponse(response.status_code,
                                response.headers, body, elapsed)
        except requests.exceptions.HTTPError as err:
            status_code_switch = {
                401: errors.AuthenticationError,
                403: errors.NotAllowedError,
                404: errors.ResourceNotFoundError,
                422: errors.ValidationError,
                429: errors.TooManyRequestsError
            }

            try:
                json_response = err.response.json()
            except ValueError:
                json_response = {}
            error_cls = status_code_switch.get(err.response.status_code,
                                               errors.ApiError)

            raise error_cls(
                request_id=err.response.headers.get(
                    constants.REQUEST_ID_HEADER),
                api_version=err.response.headers.get(
                    constants.API_VERSION_HEADER),
                http_status=err.response.status_code,
                error_type=json_response.get('error_type'),
                error_codes=json_response.get('error_codes'),
                elapsed=elapsed)
        except requests.exceptions.Timeout:
            elapsed = self._calc_elapsed_time(start)
            raise errors.ApiTimeoutError(elapsed=elapsed)
        except requests.exceptions.RequestException:
            raise IOError()

    def _calc_elapsed_time(self, start):
        return '{0:.2f}'.format((time.time() - start) * 1000)




class HTTPResponse:
    def __init__(self, status, headers, body, elapsed):
        self.status = status
        self.headers = headers
        self.body = body        # dict
        self.elapsed = elapsed  # ms

    def __str__(self):
        return '{0.status} {0.elapsed}'.format(self)



try:
    from http import HTTPStatus
except ImportError:
    # Python 3.4
    from checkout_sdk.enums import HTTPStatus

import checkout_sdk as sdk
from checkout_sdk import ApiClient, HTTPMethod, Validator, PaymentStatus
from checkout_sdk.common import ResponseDTO
from checkout_sdk.payments.responses import (
    PaymentProcessed,
    PaymentPending,
    PaymentAction
)


class PaymentsClient(ApiClient):
    def request(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            # dictionary approach - everything else is ignored
            request = args[0]
        else:
            # parameter approach
            request = kwargs

        # set defaults if attributes are missing
        self._set_payment_request_defaults(request)

        source = request.get('source')
        # 20-Sep-11, supporting destination. 
        destination = request.get('destination')
        amount = request.get('amount')
        currency = request.get('currency')
        payment_type = request.get('payment_type')
        reference = request.get('reference')

        source_type = source.get(
            'type', 'unknown') if source is not None else 'unknown'
        destination_type = destination.get( 
            'type', 'unknown') if destination is not None else 'unknown'

        self._log_info('Auth src:{}/dest:{} - {}{} - {}'.format(
            source_type, destination_type,
            amount,
            currency,
            reference if reference is not None else '<no reference>'))

        Validator.validate_transaction(
            amount=amount,
            currency=currency,
            payment_type=payment_type,
            reference=reference
        )

        # dynamic attributes
        self._set_payment_request_dynamic_attributes(request)

        http_response = self._send_http_request(
            'payments', HTTPMethod.POST, request)

        if http_response.status == HTTPStatus.ACCEPTED:
            return PaymentPending(http_response)
        else:
            return PaymentProcessed(http_response)

    def capture(self, payment_id, amount=None, reference=None, **kwargs):
        return PaymentAction(
            self._get_payment_action_response(payment_id, 'capture',
                                              amount, reference, **kwargs))

    def refund(self, payment_id, amount=None, reference=None, **kwargs):
        return PaymentAction(
            self._get_payment_action_response(payment_id, 'refund',
                                              amount, reference, **kwargs))

    def void(self, payment_id, reference=None, **kwargs):
        return PaymentAction(
            self._get_payment_action_response(payment_id, 'void',
                                              None, reference, **kwargs))

    def get(self, payment_id):
        http_response = self._get_response(payment_id, 'get')

        if http_response.body['status'] == PaymentStatus.Pending:
            return PaymentPending(http_response)
        else:
            return PaymentProcessed(http_response)

    def get_actions(self, payment_id):
        http_response = self._get_response(
            payment_id, 'get actions', '/actions')

        return [ResponseDTO(item) for item in http_response.body]

    def _get_response(self, payment_id, log_title, path=''):
        self._log_info('{} - {}'.format(log_title.capitalize(), payment_id))

        Validator.validate_id(payment_id)

        return self._send_http_request(
            'payments/{}{}'.format(payment_id, path), HTTPMethod.GET)

    def _get_payment_action_response(self, payment_id, action,
                                     amount, reference, **kwargs):
        self._log_info('{} - {} - {}'.format(
            action.capitalize(), payment_id,
            reference if reference is not None else '<no reference>'))

        Validator.validate_id(payment_id)

        request = {
            'reference': reference
        }

        if amount is not None:
            Validator.validate_transaction(amount=amount)
            request['amount'] = amount

        # add remaining properties
        request.update(kwargs)

        return self._send_http_request(
            'payments/{}/{}s'.format(payment_id, action),
            HTTPMethod.POST, request)

    def _set_payment_request_defaults(self, request):
        request['currency'] = request.get(
            'currency', sdk.default_currency)
        request['payment_type'] = request.get(
            'payment_type', sdk.default_payment_type)
        request['capture'] = request.get(
            'capture', sdk.default_capture)

    def _set_payment_request_dynamic_attributes(self, request):
        
        if request.get('source') is not None:
            self._log_debug('Source: {}'.format(request.get('source')))

            Validator.validate_complex_attribute(
                arg=request.get('source'),
                type_err_msg='Invalid payment source.',
            )
            request['source'] = Validator.validate_and_set_source_type(
                request.get('source'))

        elif request.get('destination') is not None:    
            self._log_debug('Destination: {}'.format(request.get('destination')))

            Validator.validate_complex_attribute(
                arg=request.get('destination'),
                type_err_msg='Invalid payment destination.',
            )
            request['destination'] = Validator.validate_and_set_source_type(
                request.get('destination'))
        else:
            raise ValueError("Payment Source or Destination is missing") 
        
        Validator.validate_complex_attribute(
            arg=request.get('customer'),
            type_err_msg='Invalid customer.')
        # for 3ds, due to Python name limitations,
        # we try both '3ds' and 'threeds' attribute names
        request['3ds'] = Validator.validate_and_set_dynamic_attr(
            arg=request.get(
                'threeds',
                request.get('3ds')), type_err_msg='Invalid 3DS settings.')
        request['risk'] = Validator.validate_and_set_dynamic_attr(
            arg=request.get('risk'), type_err_msg='Invalid risk settings.')



from checkout_sdk import Currency, PaymentType


class Validator:
    @classmethod
    def validate_id(cls, val):
        if val is None:
            raise ValueError('Invalid Id.')
        if not isinstance(val, str):
            raise TypeError('Id should be of string type.')

    @classmethod
    def validate_transaction(cls, amount, currency=None,
                             payment_type=None, reference=None):
        if amount is None or (isinstance(amount, int) and amount < 0):
            raise ValueError('Amount must be greater or equal to zero.')
        if not isinstance(amount, int):
            raise TypeError('Amount must be an integer.')
        if currency is not None and isinstance(currency, str) \
                and not Currency.has_value(currency):
            raise ValueError('Invalid currency.')
        if currency is not None and not isinstance(currency, str):
            raise TypeError('Currency should be a string.')
        if payment_type is not None and isinstance(payment_type, str) \
                and not PaymentType.has_value(payment_type):
            raise ValueError('Invalid payment type.')
        if payment_type is not None and not isinstance(payment_type, str):
            raise TypeError('Payment type should be a string.')
        if reference is not None and not isinstance(reference, str):
            raise TypeError('Reference must be a string.')

    @classmethod
    def validate_complex_attribute(cls, arg,
                                   type_err_msg, missing_arg_err_msg=None):
        if arg is None and missing_arg_err_msg is not None:
            raise ValueError(missing_arg_err_msg)

        if not(arg is None or isinstance(arg, dict)):
            raise TypeError(
                '{} Please provide a dictionary.'
                .format(type_err_msg))

    @classmethod
    def validate_and_set_dynamic_attr(cls, arg, type_err_msg,
                                      missing_arg_err_msg=None):
        if arg is None and missing_arg_err_msg is not None:
            raise ValueError(missing_arg_err_msg)

        if arg is None or isinstance(arg, dict):
            return arg
        elif isinstance(arg, bool):
            return {'enabled': arg}
        else:
            raise TypeError(
                '{} Please provide a dictionary or a boolean value.'
                .format(type_err_msg))

    @classmethod
    def validate_and_set_source_type(cls, source):
        source_type_switch = {
            'card': lambda: 'number' in source,
            'customer':
                lambda: 'email' in source or ('id' in source and isinstance(
                    source['id'], str) and source['id'].startswith('cus_')),
            'token': lambda: 'token' in source,
            'id':
                lambda: 'id' in source and isinstance(
                    source['id'], str) and source['id'].startswith('src_')
        }
        source_type = source.get('type', None)
        if source_type is None:
            for key in source_type_switch:
                if source_type_switch[key]():
                    # we have a match - we now set the type accordingly
                    source['type'] = key
                    return source
        else:
            return source

        # if we got here, no match found
        raise ValueError(
            'Invalid source data. Cannot determine source `type`.')



from checkout_sdk import constants
from checkout_sdk.common import ResponseDTO, HTTPResponse


class Resource(ResponseDTO):
    def __init__(self, http_response):
        if not isinstance(http_response, HTTPResponse):
            raise TypeError(
                'http_response must be a valid instance of HTTPResponse')

        super().__init__(http_response.body)
        self._response = http_response

    @property
    def http_response(self):
        """
        HTTP response with status, headers,
        JSON body and elapsed time (ms).
        """
        return self._response

    @property
    def request_id(self):
        return self._response.headers.get(constants.REQUEST_ID_HEADER)

    @property
    def api_version(self):
        return self._response.headers.get(constants.API_VERSION_HEADER)

    @property
    def links(self):
        return self._links

    @property
    def self_link(self):
        return self.get_link('self')

    def has_link(self, relation):
        return hasattr(self._links, relation)

    def get_link(self, relation):
        return getattr(self._links, relation, None)



try:
    from http import HTTPStatus
except ImportError:
    # Python 3.4
    from checkout_sdk.enums import HTTPStatus

import checkout_sdk as sdk
from checkout_sdk import ApiClient, HTTPMethod, Validator, PaymentStatus
from checkout_sdk.common import ResponseDTO
from checkout_sdk.vaults.responses import Token


class TokensClient(ApiClient):
    def request(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            # dictionary approach - everything else is ignored
            request = args[0]
        else:
            # parameter approach
            request = kwargs
 
        source_type = request.get('type')
        if(source_type == 'card'):
            number = request.get('number')
            last4 = number[-4:]
            expiry_month = request.get('expiry_month')
            expiry_year = request.get('expiry_year')
            # todo: insert validation here for cards
            self._log_debug('Token src:{} - {} {}/{}'.format(
                source_type,
                last4,
                expiry_month,
                expiry_year
            ))
        elif(source_type == 'applepay' or source_type == 'googlepay'):
            token_data = request.get('token_data')
            self._log_debug('Token src:{} - {}'.format(
                source_type,
                token_data
            ))
        

        http_response = self._send_http_request(
            'tokens', HTTPMethod.POST, request)

        return Token(http_response)